use macroquad::logging::info;
use std::f32::consts::PI;
use std::fmt::Debug;

#[derive(Debug, Clone)]
pub struct Disc {
    pub letter_map: Vec<char>,
}

impl Default for Disc {
    fn default() -> Self {
        let mut letters = ('a'..='z').into_iter().collect::<Vec<char>>();
        let mut numbers = ('0'..='9').into_iter().collect::<Vec<char>>();

        letters.append(&mut numbers);

        Self {
            letter_map: letters,
        }
    }
}

#[derive(Debug, Clone, Default)]
pub struct DecoderState {
    pub inner_disc: Disc,
    pub outer_disc: Disc,
    pub outer_disc_index: usize,
    pub outer_disc_current_angle: f32,
}

impl DecoderState {
    pub fn new(inner_map: Vec<char>, outer_map: Vec<char>) -> Self {
        Self {
            inner_disc: Disc {
                letter_map: inner_map,
            },
            outer_disc: Disc {
                letter_map: outer_map,
            },
            ..Default::default()
        }
    }

    pub fn set_index(&mut self, index: usize) {
        self.outer_disc_index = index;
    }

    pub fn change_index(&mut self, diff: i32) {
        let new_index = (((self.outer_disc_index as i32) + diff)
            .rem_euclid(self.outer_disc.letter_map.len() as i32)) as usize;

        info!("New Index: {}", new_index);
        self.set_index(new_index);
    }

    pub fn get_angle_per_letter(&self) -> f32 {
        (2.0 * PI) / (self.outer_disc.letter_map.len() as f32)
    }

    pub fn get_current_target_angle(&self) -> f32 {
        (2.0 * PI) - self.get_angle_per_letter() * (self.outer_disc_index as f32)
    }

    pub fn get_outer_letter_at_index(&self, index: usize) -> char {
        let index = (index as i32 + self.outer_disc_index as i32)
            .rem_euclid(self.outer_disc.letter_map.len() as i32) as usize;
        self.outer_disc.letter_map[index]
    }

    pub fn get_index_from_inner_letter(&self, c: char) -> usize {
        self.inner_disc
            .letter_map
            .iter()
            .position(|n| *n == c)
            .unwrap()
    }

    pub fn get_index_from_outer_letter(&self, c: char) -> usize {
        let norm_index = self
            .outer_disc
            .letter_map
            .iter()
            .position(|n| *n == c)
            .unwrap();

        let index = (norm_index as i32 - self.outer_disc_index as i32)
            .rem_euclid(self.outer_disc.letter_map.len() as i32) as usize;

        return index;
    }

    pub fn get_inner_letter_at_index(&self, index: usize) -> char {
        self.inner_disc.letter_map[index]
    }

    pub fn encode(&mut self, in_char: char) -> char {
        let index = self.get_index_from_inner_letter(in_char);

        let char = self.get_outer_letter_at_index(index);

        return char;
    }

    pub fn decode(&mut self, in_char: char) -> char {
        let index = self.get_index_from_outer_letter(in_char);

        let char = self.get_inner_letter_at_index(index);

        return char;
    }
}
