mod disc;
mod config;

use crate::disc::{DecoderState, Disc};
use macroquad::prelude::*;
use std::f32::consts::PI;
use std::string::ToString;
use crate::config::Config;

const SHADOW: Color = Color::new(0.25, 0.25, 0.25, 0.5);

const GAME_NAME: &str = "dial";

pub struct Context {
    decoder_state: DecoderState,
    input_state: DecoderState,
    mode: Mode,
}

#[derive(Eq, PartialEq)]
pub enum Mode {
    Encode,
    Decode,
}

pub fn conf() -> Conf {
    Conf {
        window_title: GAME_NAME.to_string(),
        high_dpi: true,
        fullscreen: true,
        ..Default::default()
    }
}

pub fn game_init() {
    info!("Starting {}", GAME_NAME);
}

pub fn rotate_vec2(vec: Vec2, angle: f32) -> Vec2 {
    Mat2::from_angle(angle) * vec
}

pub fn draw_display(pos: Vec2, size: Vec2, text: &str) {
    let font_size = size.y;
    let text_center = get_text_center(text, None, font_size as u16, 1.0, 0.0);

    draw_rectangle_ex(
        pos.x,
        pos.y,
        size.x,
        size.y,
        DrawRectangleParams {
            rotation: 0.0,
            color: BLACK,
            ..Default::default()
        },
    );

    let text_pos = pos + size * 0.5 - text_center;
    draw_text(text, text_pos.x, text_pos.y, font_size, SKYBLUE);
}

pub fn draw_label(pos: Vec2, size: Vec2, text: &str) {
    let font_size = size.y;
    let text_center = get_text_center(text, None, font_size as u16, 1.0, 0.0);

    draw_rectangle_ex(
        pos.x,
        pos.y,
        size.x,
        size.y,
        DrawRectangleParams {
            rotation: 0.0,
            color: BLACK,
            ..Default::default()
        },
    );

    let text_pos = pos + size * 0.5 - text_center;
    draw_text(text, text_pos.x, text_pos.y, font_size, WHITE);
}

pub fn draw_disc(
    disc: &Disc,
    base_pos: Vec2,
    text_radius: f32,
    radius: f32,
    start_angle: f32,
    color: Color,
) {
    draw_circle(
        base_pos.x + radius * 0.05,
        base_pos.y + radius * 0.05,
        radius,
        SHADOW,
    );
    draw_circle(base_pos.x, base_pos.y, radius, color);
    draw_circle(base_pos.x, base_pos.y, radius * 0.10, BLACK);
    draw_circle_lines(
        base_pos.x,
        base_pos.y,
        radius - radius * 0.01,
        radius * 0.025,
        BLACK,
    );

    let len = disc.letter_map.len();

    let angle_per_letter = (2.0 * PI) / (len as f32);

    let mut text_pos = Vec2::new(0.0, -text_radius);
    let mut line_pos = Vec2::new(0.0, radius);

    let font_size = (radius * 0.15) as u16;

    let mut angle = start_angle;

    text_pos = rotate_vec2(text_pos, start_angle);
    line_pos = rotate_vec2(line_pos, start_angle);

    for (ndx, letter) in disc.letter_map.iter().enumerate() {
        let text_center = get_text_center(letter.to_string().as_str(), None, font_size, 1.0, angle);
        let letter_pos = text_pos + base_pos - text_center;

        let mut line_pos_1 = rotate_vec2(line_pos, -angle_per_letter * 0.5);
        line_pos_1 = line_pos_1 + base_pos;
        let mut line_pos_2 = rotate_vec2(line_pos, angle_per_letter * 0.5);
        line_pos_2 = line_pos_2 + base_pos;

        draw_line(
            base_pos.x,
            base_pos.y,
            line_pos_1.x,
            line_pos_1.y,
            2.0,
            BLACK,
        );
        draw_line(
            base_pos.x,
            base_pos.y,
            line_pos_2.x,
            line_pos_2.y,
            2.0,
            BLACK,
        );

        if ndx == 0 {
            draw_circle(
                text_pos.x + base_pos.x,
                text_pos.y + base_pos.y,
                radius * 0.05,
                GRAY,
            );
        }

        draw_text_ex(
            letter.to_string().as_str(),
            letter_pos.x,
            letter_pos.y,
            TextParams {
                font_size,
                color: WHITE,
                rotation: angle,
                ..Default::default()
            },
        );

        text_pos = rotate_vec2(text_pos, angle_per_letter);
        line_pos = rotate_vec2(line_pos, angle_per_letter);
        angle += angle_per_letter;
    }
}

pub fn draw_switch(pos: Vec2, size: Vec2, mut state: bool) -> bool {
    draw_circle(pos.x, pos.y, size.x * 0.5, BLACK);

    if is_mouse_in_area(pos - size * 0.5, size) {
        state = !state;
    }

    let rot = if state { 180.0_f32.to_radians() } else { 0.0 };

    draw_rectangle_ex(
        pos.x + (size.x * 0.07),
        pos.y + (size.y * 0.1),
        size.x * 0.33,
        size.y,
        DrawRectangleParams {
            rotation: rot,
            color: SHADOW,
            offset: Vec2::new(0.5, 0.0),
        },
    );

    draw_rectangle_ex(
        pos.x,
        pos.y,
        size.x * 0.33,
        size.y,
        DrawRectangleParams {
            rotation: rot,
            color: DARKBLUE,
            offset: Vec2::new(0.5, 0.0),
        },
    );

    draw_rectangle_lines_ex(
        pos.x,
        pos.y,
        size.x * 0.33,
        size.y,
        size.x * 0.025,
        DrawRectangleParams {
            rotation: rot,
            color: BLACK,
            offset: Vec2::new(0.5, 0.0),
        },
    );

    return state;
}

pub fn is_mouse_in_area(pos: Vec2, size: Vec2) -> bool {
    //draw_rectangle_lines(pos.x, pos.y, size.x, size.y, 5.0, RED);

    if is_mouse_button_pressed(MouseButton::Left) {
        let lower_bound = pos;
        let upper_bound = pos + size;
        let mouse_pos: Vec2 = mouse_position().into();

        let lower_bound_check = lower_bound.x <= mouse_pos.x && lower_bound.y <= mouse_pos.y;
        let upper_bound_check = upper_bound.x >= mouse_pos.x && upper_bound.y >= mouse_pos.y;

        if lower_bound_check && upper_bound_check {
            return true;
        }
    }

    false
}

pub fn draw_device(ctx: &mut Context) {
    let screen_height = screen_height();
    let screen_width = screen_width();

    let width = screen_width * 0.33;
    let height = screen_height * 0.9;

    let dev_x = screen_width * 0.5 - width * 0.5;
    let dev_y = screen_height * 0.5 - height * 0.5;

    let size = Vec2::new(width, height);
    let base_pos = Vec2::new(dev_x, dev_y);

    input_dial(&mut ctx.input_state, base_pos, size);

    draw_rectangle_ex(
        dev_x,
        dev_y,
        width,
        height,
        DrawRectangleParams {
            offset: Default::default(),
            rotation: 0.0,
            color: DARKGREEN,
        },
    );

    draw_rectangle_lines(dev_x, dev_y, width, height, dev_x * 0.005, BLACK);

    draw_rectangle_ex(
        dev_x + width * 0.02,
        dev_y + height * 0.015,
        width * 0.68,
        height * 0.05,
        DrawRectangleParams {
            offset: Default::default(),
            rotation: 0.0,
            color: LIGHTGRAY,
        },
    );

    draw_text(
        "PROPERTY OF Foghaven Guard",
        dev_x + width * 0.02,
        dev_y + height * 0.05,
        width * 0.06,
        RED,
    );

    input_panel(ctx, base_pos, size);

    let next_button_pos = Vec2::new(size.x * 0.9, size.y * 0.4) + base_pos;
    let next_button_size = Vec2::new(size.x * 0.1, size.x * 0.1);

    if is_mouse_in_area(next_button_pos - next_button_size * 0.5, next_button_size) {
        ctx.decoder_state.change_index(1);
    }

    draw_circle(
        next_button_pos.x + next_button_size.x * 0.05,
        next_button_pos.y + next_button_size.y * 0.05,
        next_button_size.x * 0.5,
        SHADOW,
    );
    draw_circle(
        next_button_pos.x,
        next_button_pos.y,
        next_button_size.x * 0.5,
        RED,
    );

    let base_pos = Vec2::new(dev_x + width * 0.5, dev_y + height * 0.66);

    if ctx.decoder_state.get_current_target_angle() - ctx.decoder_state.outer_disc_current_angle
        > f32::EPSILON
    {
        ctx.decoder_state.outer_disc_current_angle += 1.0_f32.to_radians() * (get_time() as f32);
    } else {
        ctx.decoder_state.outer_disc_current_angle = ctx.decoder_state.get_current_target_angle();
    }

    draw_disc(
        &ctx.decoder_state.outer_disc,
        base_pos,
        width * 0.40,
        width * 0.45,
        ctx.decoder_state.outer_disc_current_angle,
        BROWN,
    );

    draw_disc(
        &ctx.decoder_state.inner_disc,
        base_pos,
        width * 0.20,
        width * 0.25,
        0.0,
        GOLD,
    );

    let surplus_sticker_pos = Vec2::new(dev_x + width * 0.75, dev_y + height * 0.01);
    let surplus_sticker_size = Vec2::new(width * 0.24, height * 0.06);
    draw_rectangle_ex(
        surplus_sticker_pos.x + surplus_sticker_size.x * 0.02,
        surplus_sticker_pos.y + surplus_sticker_size.x * 0.02,
        surplus_sticker_size.x,
        surplus_sticker_size.y,
        DrawRectangleParams {
            offset: Default::default(),
            rotation: 10.0_f32.to_radians(),
            color: SHADOW,
        },
    );

    draw_rectangle_ex(
        surplus_sticker_pos.x,
        surplus_sticker_pos.y,
        surplus_sticker_size.x,
        surplus_sticker_size.y,
        DrawRectangleParams {
            offset: Default::default(),
            rotation: 10.0_f32.to_radians(),
            color: Color::new(0.25, 0.25, 0.25, 1.0),
        },
    );
    draw_text_ex(
        "(SURPLUS)",
        dev_x + width * 0.75,
        dev_y + height * 0.05,
        TextParams {
            font: None,
            font_size: (dev_x * 0.055) as u16,
            font_scale: 1.0,
            font_scale_aspect: 1.0,
            rotation: 10.0f32.to_radians(),
            color: RED,
        },
    );
}

fn input_dial(input_disc: &mut DecoderState, base: Vec2, size: Vec2) {
    let dial_pos = Vec2::new(base.x + size.x * 0.15, base.y + size.y * 0.3);

    draw_disc(
        &input_disc.outer_disc,
        dial_pos,
        size.y * 0.15,
        size.y * 0.2,
        input_disc.outer_disc_current_angle,
        DARKGREEN,
    );

    let marker_pos = dial_pos + Vec2::new(-size.y * 0.11, 0.0);
    draw_circle(marker_pos.x, marker_pos.y, size.y * 0.005, RED);

    if input_disc.get_current_target_angle() - input_disc.outer_disc_current_angle > f32::EPSILON {
        input_disc.outer_disc_current_angle += 1.0_f32.to_radians() * (get_time() as f32);
    } else {
        input_disc.outer_disc_current_angle = input_disc.get_current_target_angle();
    }

    let click_area = Vec2::new(size.y * 0.1, size.y * 0.15);
    let upper_click_pos = dial_pos - Vec2::new(size.y * 0.1, 0.0) - click_area;
    let lower_click_pos = dial_pos - Vec2::new(size.y * 0.1, -click_area.y) - click_area;

    if is_mouse_in_area(upper_click_pos, click_area) {
        input_disc.change_index(1);
    } else if is_mouse_in_area(lower_click_pos, click_area) {
        input_disc.change_index(-1);
    }
}

fn input_panel(ctx: &mut Context, base: Vec2, size: Vec2) {
    let border_pos = Vec2::new(base.x + size.x * 0.05, base.y + size.y * 0.075);
    let border_size = Vec2::new(size.x * 0.9, size.y * 0.25);

    draw_rectangle_ex(
        border_pos.x + border_size.x * 0.01,
        border_pos.y + border_size.y * 0.02,
        border_size.x,
        border_size.y,
        DrawRectangleParams {
            offset: Default::default(),
            rotation: 0.0,
            color: SHADOW,
        },
    );

    draw_rectangle_ex(
        border_pos.x,
        border_pos.y,
        border_size.x,
        border_size.y,
        DrawRectangleParams {
            offset: Default::default(),
            rotation: 0.0,
            color: GRAY,
        },
    );

    draw_rectangle_lines(
        border_pos.x,
        border_pos.y,
        border_size.x,
        border_size.y,
        border_size.x * 0.005,
        BLACK,
    );

    let input_letter_disp_loc = border_pos + Vec2::new(border_size.x * 0.05, border_size.x * 0.05);
    let input_letter_size = Vec2::new(border_size.x * 0.10, border_size.y * 0.3);

    let input_letter = ctx
        .input_state
        .get_outer_letter_at_index(ctx.input_state.outer_disc.letter_map.len() * 3 / 4);
    draw_display(
        input_letter_disp_loc,
        input_letter_size,
        input_letter.to_string().as_str(),
    );

    let input_label_pos = input_letter_disp_loc
        + Vec2::new(
            input_letter_size.x + border_size.x * 0.05,
            input_letter_size.y * 0.25,
        );
    let input_label_size = Vec2::new(border_size.x * 0.2, border_size.y * 0.15);
    draw_label(input_label_pos, input_label_size, "INPUT");

    let index_disp_loc =
        input_letter_disp_loc + Vec2::new(0.0, input_letter_size.y + border_size.x * 0.05);
    draw_display(
        index_disp_loc,
        input_letter_size,
        ctx.decoder_state.outer_disc_index.to_string().as_str(),
    );

    let index_label_pos = index_disp_loc
        + Vec2::new(
            input_letter_size.x + border_size.x * 0.05,
            input_letter_size.y * 0.25,
        );

    draw_label(index_label_pos, input_label_size, "INDEX");

    let output_disp_loc = border_pos + Vec2::new(border_size.x * 0.55, border_size.x * 0.05);
    let output_disp_size = Vec2::new(border_size.x * 0.33, border_size.y * 0.6);

    let output_disp = if ctx.mode == Mode::Encode {
        ctx.decoder_state.encode(input_letter)
    } else {
        ctx.decoder_state.decode(input_letter)
    };

    draw_display(
        output_disp_loc,
        output_disp_size,
        output_disp.to_string().as_str(),
    );

    let output_label_size = input_label_size * 1.1;

    let output_disp_label_loc = output_disp_loc
        + Vec2::new(
            output_disp_size.x / 2.0 - output_label_size.x / 2.0,
            output_disp_size.y + size.y * 0.01,
        );

    draw_label(output_disp_label_loc, output_label_size, "OUTPUT");

    let switch_pos = border_pos + border_size * 0.5;
    let switch_size = Vec2::new(border_size.x * 0.05, border_size.y * 0.25);
    draw_label(
        switch_pos + Vec2::new(-border_size.x * 0.05 * 0.5, switch_size.y),
        Vec2::new(border_size.x * 0.05, border_size.x * 0.05),
        "D",
    );
    draw_label(
        switch_pos + Vec2::new(-border_size.x * 0.05 * 0.5, -switch_size.y * 1.5),
        Vec2::new(border_size.x * 0.05, border_size.x * 0.05),
        "E",
    );

    if draw_switch(switch_pos, switch_size, ctx.mode == Mode::Encode) {
        if ctx.mode != Mode::Encode {
            info!("Mode set to encode!");
            ctx.mode = Mode::Encode
        }
    } else {
        if ctx.mode != Mode::Decode {
            info!("Mode set to decode!");
            ctx.mode = Mode::Decode
        }
    }
}

#[macroquad::main(conf)]
async fn main() {
    game_init();

    let config = load_file("config.toml").await.unwrap();
    let config_str = String::from_utf8(config).unwrap();

    let config: Config = toml::from_str(config_str.as_str()).unwrap();

    let decoder = DecoderState::new(config.get_inner_disc(), config.get_outer_disc());
    let input_disc = DecoderState::default();

    let mut context = Context {
        decoder_state: decoder,
        input_state: input_disc,
        mode: Mode::Decode,
    };

    context
        .input_state
        .set_index(context.input_state.outer_disc.letter_map.len() / 4);

    loop {
        clear_background(WHITE);

        draw_device(&mut context);

        next_frame().await;
    }
}
